"""
This module contains test for the radish EBNF gherkin parser.
"""

from pathlib import Path

import pytest

from radish.models import (
    ConstantTag,
    DefaultRule,
    PreconditionTag,
    Scenario,
    ScenarioLoop,
    ScenarioOutline,
)
from radish.parser import FeatureFileParser
from radish.parser.errors import (
    RadishFirstStepMustUseFirstLevelKeyword,
    RadishLanguageNotFound,
    RadishMisplacedBackground,
    RadishMissingFeatureShortDescription,
    RadishMissingRuleShortDescription,
    RadishMissingScenarioShortDescription,
    RadishMultipleBackgrounds,
    RadishScenarioLoopInvalidIterationsValue,
    RadishScenarioLoopMissingIterations,
    RadishScenarioOutlineExamplesInconsistentCellCount,
    RadishScenarioOutlineExamplesMissingClosingVBar,
    RadishScenarioOutlineExamplesMissingOpeningVBar,
    RadishScenarioOutlineWithoutExamples,
    RadishStepDataTableInconsistentCellCount,
    RadishStepDataTableMissingClosingVBar,
    RadishStepDocStringNotClosed,
    RadishStepDoesNotStartWithKeyword,
)

FEATURE_FILES_DIR = Path(__file__).parent / "features"


@pytest.fixture(name="parser")
def setup_default_featurefileparser(mocker):
    parser = FeatureFileParser()
    parser._resolve_preconditions = mocker.MagicMock()
    return parser


def test_parse_empty_feature_file(parser):
    """The parser should parse an empty Feature File"""
    # given
    feature_file = ""

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast is None


def test_parse_short_description_from_empty_feature(parser):
    """The parser should parse the short description of an empty Feature"""
    # given
    feature_file = """
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.short_description == "My Feature"


def test_parse_fail_no_short_description_from_feature(parser):
    """The parser should be fail to parse a Feature without a short description"""
    # given
    feature_file = """
        Feature:
    """

    # then
    with pytest.raises(RadishMissingFeatureShortDescription):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.xfail(
    reason="Not implemented yet. Currently the second Feature is parsed as description"
)
def test_parse_fail_multiple_features_in_file(parser):
    """The parser should fail to parse multiple Features in a single Feature File"""
    # given
    feature_file = """
        Feature: My first Feature
        Feature: My second Feature
    """

    # then
    # with pytest.raises(RadishMissingFeatureShortDescription):
    # when
    parser.parse_contents(None, feature_file)


def test_parse_short_description_and_description_from_feature(parser):
    """The parser should parse the short description and the description of a Feature"""
    # given
    feature_file = """
        Feature: My Feature
            This is the description
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.short_description == "My Feature"
    assert ast.description == ["This is the description"]


def test_parse_multiline_description_from_feature(parser):
    """The parser should parse a multiline description of a Feature"""
    # given
    feature_file = """
        Feature: My Feature
            This is the description
            on multiple lines.
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.description == ["This is the description", "on multiple lines."]


def test_parse_multiline_description_with_blanklines_from_feature(parser):
    """
    The parser should parse a multiline description
    containing blank lines of a Feature
    """
    # given
    feature_file = """
        Feature: My Feature
            This is the description
            on multiple lines.

            It even contains blank lines
            which is great for readability.
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.description == [
        "This is the description",
        "on multiple lines." "",
        "It even contains blank lines",
        "which is great for readability.",
    ]


def test_parse_tag_from_a_feature(parser):
    """The parser should parse a Tag from a Feature"""
    # given
    feature_file = """
        @tag-a
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.tags) == 1
    assert ast.tags[0].name == "tag-a"


def test_parse_tags_on_multiple_lines_from_a_feature(parser):
    """The parser should parse multiple Tags on multiple lines from a Feature"""
    # given
    feature_file = """
        @tag-a
        @tag-b
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.tags) == 2
    assert ast.tags[0].name == "tag-a"
    assert ast.tags[1].name == "tag-b"


def test_parse_tags_on_same_line_from_a_feature(parser):
    """The parser should parse multiple Tags on the same line from a Feature"""
    # given
    feature_file = """
        @tag-a @tag-b
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.tags) == 2
    assert ast.tags[0].name == "tag-a"
    assert ast.tags[1].name == "tag-b"


def test_parse_tags_on_multiple_and_same_line_from_a_feature(parser):
    """
    The parser should parse multiple Tags on multiple and
    the same line from a Feature
    """
    # given
    feature_file = """
        @tag-a
        @tag-b @tag-c
        @tag-d
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.tags) == 4
    assert ast.tags[0].name == "tag-a"
    assert ast.tags[1].name == "tag-b"
    assert ast.tags[2].name == "tag-c"
    assert ast.tags[3].name == "tag-d"


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Rule: My Rule
    """,
        pytest.param(
            """
        Feature: My Feature
        Rule: My Rule
    """,
            marks=pytest.mark.xfail(
                reason="A Feature without description must have a blank line before the first block"
            ),
        ),
    ],
    ids=[
        "with a blank line between the Feature and the Rule",
        "without a blank line between the Feature and the Rule",
    ],
)
def test_parse_single_empty_rule(parser, feature_file):
    """The parser should parse a single empty Rule"""
    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules) == 1
    assert ast.rules[0].short_description == "My Rule"


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Rule: My first Rule
            Rule: My second Rule
    """,
        """
        Feature: My Feature

            Rule: My first Rule

            Rule: My second Rule
    """,
    ],
    ids=[
        "without a blank line between the Rules",
        "With a blank line between the Rules",
    ],
)
def test_parse_multiple_empty_rules(parser, feature_file):
    """The parser should parse multiple empty Rules"""
    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules) == 2
    assert ast.rules[0].short_description == "My first Rule"
    assert ast.rules[1].short_description == "My second Rule"


@pytest.mark.xfail(reason="Not Supported Yet")
def test_parse_rule_description(parser):
    """The parser should parse the Rule description"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My Rule
                This is the Rule description
                on multiple lines even.
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].description == [
        "This is the Rule description",
        "on multiple lines even.",
    ]


def test_parse_fail_no_short_description_in_rule(parser):
    """The parser should fail to parse a Rule without a short description"""
    # given
    feature_file = """
        Feature: My Feature

            Rule:
    """
    # then
    with pytest.raises(RadishMissingRuleShortDescription):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_single_empty_scenario_outside_rule(parser):
    """The parser should parse a single empty Scenario outside any Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert ast.rules[0].scenarios[0].short_description == "My Scenario"


def test_parse_example_synonym_for_scenario(parser):
    """The parser should recognize Example as a synonym for Scenario"""
    # given
    feature_file = """
        Feature: My Feature

            Example: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], Scenario)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario"


def test_parse_create_default_rule_for_scenario_outside_rule(parser):
    """The parser should create a DefaultRule for a Scenario outside any Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules) == 1
    assert isinstance(ast.rules[0], DefaultRule)
    assert ast.rules[0].short_description is None
    assert ast.rules[0].line == 4
    assert len(ast.rules[0].scenarios) == 1


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario: My first Scenario
            Scenario: My second Scenario
    """,
        """
        Feature: My Feature

            Scenario: My first Scenario

            Scenario: My second Scenario
    """,
    ],
    ids=[
        "without a blank line between the Scenarios",
        "With a blank line between the Scenario",
    ],
)
def test_parse_multiple_empty_scenarios_outside_rule(parser, feature_file):
    """The parser should parse multiple empty Scenarios outside any Rule"""
    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 2
    assert ast.rules[0].scenarios[0].short_description == "My first Scenario"
    assert ast.rules[0].scenarios[1].short_description == "My second Scenario"


def test_parse_fail_no_short_description_in_scenario(parser):
    """The parser should fail to parse a Scenario without a short description"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario:
    """
    # then
    with pytest.raises(RadishMissingScenarioShortDescription):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_single_empty_scenario_inside_rule(parser):
    """The parser should parse a single empty Scenario inside a Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My Rule
                Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My Rule"
    assert ast.rules[0].scenarios[0].short_description == "My Scenario"


def test_parse_multiple_empty_scenario_inside_rule(parser):
    """The parser should parse multiple empty Scenarios inside a Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My Rule
                Scenario: My first Scenario
                Scenario: My second Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My Rule"
    assert ast.rules[0].scenarios[0].short_description == "My first Scenario"
    assert ast.rules[0].scenarios[1].short_description == "My second Scenario"


def test_parse_single_empty_scenario_inside_multiple_rules(parser):
    """The parser should parse a single empty Scenario inside multiple Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My first Rule
                Scenario: My Scenario

            Rule: My second Rule
                Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My first Rule"
    assert ast.rules[0].scenarios[0].short_description == "My Scenario"
    assert ast.rules[1].short_description == "My second Rule"
    assert ast.rules[1].scenarios[0].short_description == "My Scenario"


def test_parse_multiple_empty_scenarios_inside_multiple_rules(parser):
    """The parser should parse multiple empty Scenarios inside multiple Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My first Rule
                Scenario: My first Scenario
                Scenario: My second Scenario

            Rule: My second Rule
                Scenario: My first Scenario
                Scenario: My second Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My first Rule"
    assert ast.rules[0].scenarios[0].short_description == "My first Scenario"
    assert ast.rules[0].scenarios[1].short_description == "My second Scenario"
    assert ast.rules[1].short_description == "My second Rule"
    assert ast.rules[1].scenarios[0].short_description == "My first Scenario"
    assert ast.rules[1].scenarios[1].short_description == "My second Scenario"


def test_parse_defaultrule_scenario_and_named_rule_scenario(parser):
    """The parser should parse a Scenario outside any Rules followed by a Scenario inside a Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario

            Rule: My Rule
                Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert isinstance(ast.rules[0], DefaultRule)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario"
    assert ast.rules[1].short_description == "My Rule"
    assert ast.rules[1].scenarios[0].short_description == "My Scenario"


def test_parse_tag_from_a_scenario(parser):
    """The parser should parse a Tag from a Scenario"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 1
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"


def test_parse_tags_on_multiple_lines_from_a_scenario(parser):
    """The parser should parse multiple Tags on multiple lines from a Scenario"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            @tag-b
            Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 2
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"
    assert ast.rules[0].scenarios[0].tags[1].name == "tag-b"


def test_parse_empty_background_without_short_description(parser):
    """The parser should parse an empty Background without a short description"""
    # given
    feature_file = """
        Feature: My Feature

            Background:
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.background is not None


def test_parse_empty_background_with_short_description(parser):
    """The parser should parse an empty Background with a short description"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.background.short_description == "My Background"


def test_parse_background_has_assigned_feature(parser):
    """The parser should assign the Feature to the Background"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.background.feature is not None


@pytest.mark.xfail(
    reason="RadishMultipleBackgrounds is not matched. Probably a bug in lark?"
)
def test_parse_fail_multiple_backgrounds(parser):
    """The parser should fail if multiple Backgrounds exist"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My first Background
            Background: My second Background
    """

    # when
    with pytest.raises(RadishMultipleBackgrounds):
        # when
        parser.parse_contents(None, feature_file)


def test_parser_assign_background_to_single_scenario_outside_rule(parser):
    """The parser should assign a parsed Background to a single Scenario outside any Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background

            Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].background is not None


def test_parser_assign_background_to_multiple_scenarios_outside_rule(parser):
    """The parser should assign a parsed Background to a every Scenario outside any Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background

            Scenario: My first Scenario
            Scenario: My second Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].background is not None
    assert ast.rules[0].scenarios[1].background is not None

    assert (
        ast.rules[0].scenarios[0].background is not ast.rules[0].scenarios[1].background
    )


def test_parser_assign_background_to_single_scenario_inside_rule(parser):
    """The parser should assign a parsed Background to a single Scenario inside a Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background

            Rule: My Rule
                Scenario: My Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].background is not None


def test_parser_assign_background_to_every_scenario_inside_rules(parser):
    """The parser should assign a parsed Background to a every Scenarios inside Rules"""
    # given
    feature_file = """
        Feature: My Feature

            Background: My Background

            Rule: My first Rule
                Scenario: My first Scenario
                Scenario: My second Scenario

            Rule: My second Rule
                Scenario: My first Scenario
                Scenario: My second Scenario
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].background is not None
    assert ast.rules[0].scenarios[1].background is not None
    assert ast.rules[1].scenarios[0].background is not None
    assert ast.rules[1].scenarios[1].background is not None

    assert (
        ast.rules[0].scenarios[0].background is not ast.rules[0].scenarios[1].background
    )
    assert (
        ast.rules[0].scenarios[1].background is not ast.rules[1].scenarios[0].background
    )
    assert (
        ast.rules[1].scenarios[0].background is not ast.rules[1].scenarios[1].background
    )


@pytest.mark.xfail(
    reason="RadishMisplacedBackground is not matched. Probably a bug in lark?"
)
def test_parse_fail_misplaced_background(parser):
    """The parser should fail for misplaced Backgrounds"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario

            Background: My Background
    """

    # when
    with pytest.raises(RadishMisplacedBackground):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize("keyword", ["Given", "When", "Then"])
def test_parse_single_scenario_with_single_step(parser, keyword):
    """The parser should parse a single Scenario with a single Step"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                {} there is a Step
    """.format(
        keyword
    )

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == keyword
    assert ast.rules[0].scenarios[0].steps[0].used_keyword == keyword
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a Step"


def test_parse_fail_step_with_invalid_keyword(parser):
    """The parser should fail to parse a Step with an invalid keyword"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Foo there is a Step
    """

    # then
    with pytest.raises(RadishStepDoesNotStartWithKeyword):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario: My Scenario
                And there is a Step
    """,
        """
        Feature: My Feature

            Scenario: My first Scenario
                Given there is a Step

            Scenario: My second Scenario
                And there is a Step
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                But there is a Step
    """,
        """
        Feature: My Feature

            Scenario: My first Scenario
                Given there is a Step

            Scenario: My second Scenario
                But there is a Step
    """,
    ],
    ids=[
        "'And' in first Scenario",
        "'And' in second Scenario",
        "'But' in first Scenario",
        "'But' in second Scenario",
    ],
)
def test_parse_fail_first_step_no_first_level_keyword(parser, feature_file):
    """The parser should fail to parse if the first Step has no first level keyword"""
    # then
    with pytest.raises(RadishFirstStepMustUseFirstLevelKeyword):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_second_level_keyword_assigned_correct_first_level_keyword(parser):
    """
    The parser should assign the correct first level keyword to a Step with a second level keyword
    """
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a Step
                And there is a Step
                When there is a Step
                And there is a Step
                Then there is a Step
                And there is a Step
                But there is a Step
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    steps = ast.rules[0].scenarios[0].steps
    assert steps[0].keyword == "Given"
    assert steps[0].used_keyword == "Given"
    assert steps[1].keyword == "Given"
    assert steps[1].used_keyword == "And"
    assert steps[2].keyword == "When"
    assert steps[2].used_keyword == "When"
    assert steps[3].keyword == "When"
    assert steps[3].used_keyword == "And"
    assert steps[4].keyword == "Then"
    assert steps[4].used_keyword == "Then"
    assert steps[5].keyword == "Then"
    assert steps[5].used_keyword == "And"
    assert steps[6].keyword == "Then"
    assert steps[6].used_keyword == "But"


def test_parse_keyword_context_reset_after_scenario(parser):
    """The parser should reset the frist level keyword context for each new Scenario"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My first Scenario
                Given there is a Step
                And there is a Step

            Scenario: My second Scenario
                When there is a Step
                And there is a Step
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    first_scenario_steps = ast.rules[0].scenarios[0].steps
    second_scenario_steps = ast.rules[0].scenarios[1].steps
    assert first_scenario_steps[0].keyword == "Given"
    assert first_scenario_steps[1].keyword == "Given"
    assert second_scenario_steps[0].keyword == "When"
    assert second_scenario_steps[1].keyword == "When"


def test_parse_single_scenario_with_multiple_steps(parser):
    """The parser should parse a single Scenario with multiple Steps"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                When there is an action
                Then there is an assertion
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 3
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[0].steps[1].text == "there is an action"
    assert ast.rules[0].scenarios[0].steps[2].keyword == "Then"
    assert ast.rules[0].scenarios[0].steps[2].text == "there is an assertion"


def test_parse_scenario_with_multiple_steps_separated_by_blanklines(parser):
    """The parser should parse a Scenario with multiple Steps separated by blank lines"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup

                When there is an action

                Then there is an assertion
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 3
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[0].steps[1].text == "there is an action"
    assert ast.rules[0].scenarios[0].steps[2].keyword == "Then"
    assert ast.rules[0].scenarios[0].steps[2].text == "there is an assertion"


def test_parse_background_with_multiple_steps(parser):
    """The parser should parse a Background with multiple Steps"""
    # given
    feature_file = """
        Feature: My Feature

            Background:
                Given there is a setup
                When there is an action
                Then there is an assertion
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.background.steps) == 3
    assert ast.background.steps[0].keyword == "Given"
    assert ast.background.steps[0].text == "there is a setup"
    assert ast.background.steps[1].keyword == "When"
    assert ast.background.steps[1].text == "there is an action"
    assert ast.background.steps[2].keyword == "Then"
    assert ast.background.steps[2].text == "there is an assertion"


def test_parse_multiple_scenarios_with_multiple_steps(parser):
    """The parser should parse multiple Scenarios with multiple Steps"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My first Scenario
                Given there is a setup
                When there is an action
                Then there is an assertion

            Scenario: My second Scenario
                Given there is a setup
                When there is an action
                Then there is an assertion
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 2
    assert len(ast.rules[0].scenarios[0].steps) == 3
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[0].steps[1].text == "there is an action"
    assert ast.rules[0].scenarios[0].steps[2].keyword == "Then"
    assert ast.rules[0].scenarios[0].steps[2].text == "there is an assertion"
    assert len(ast.rules[0].scenarios[1].steps) == 3
    assert ast.rules[0].scenarios[1].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[1].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[1].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[1].steps[1].text == "there is an action"
    assert ast.rules[0].scenarios[1].steps[2].keyword == "Then"
    assert ast.rules[0].scenarios[1].steps[2].text == "there is an assertion"


def test_parse_step_with_single_line_doc_string(parser):
    """The parser should parse a Step with a single-line Doc String"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == "My Doc String\n"


def test_parse_step_with_single_line_doc_string_consecutive_step(parser):
    """The parser should parse a Step with a single-line Doc String and a following Step"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                \"\"\"
                When there is a Step
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 2
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == "My Doc String\n"
    assert ast.rules[0].scenarios[0].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[0].steps[1].text == "there is a Step"
    assert ast.rules[0].scenarios[0].steps[1].doc_string is None


def test_parse_step_with_multi_line_doc_string(parser):
    """The parser should parse a Step with a multi-line Doc String"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                Another line in the Doc String
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == (
        "My Doc String\nAnother line in the Doc String\n"
    )


def test_parse_step_with_multi_line_doc_string_with_blank_lines(parser):
    """The parser should parse a Step with a multi-line Doc String containing blank lines"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                Before blank lines


                After blank lines
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == (
        "Before blank lines\n" "\n" "\n" "After blank lines\n"
    )


def test_parse_dedent_doc_string(parser):
    """The parser should dedent the Doc String to the common leading whitespaces"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                    which is somewhat indented
                which is awesome
                    and useful
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == (
        "My Doc String\n"
        "    which is somewhat indented\n"
        "which is awesome\n"
        "    and useful\n"
    )


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                When I do something
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                \"
                When I do something
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                \"\"
                When I do something
    """,
    ],
    ids=[
        "not closed at all",
        "only one consecutive double quotes for closing",
        "only two consecutive double quotes for closing",
    ],
)
def test_parse_fail_doc_string_not_closed(parser, feature_file):
    """The parser should fail to parse a Doc String which isn't closed properly"""
    # then
    with pytest.raises(RadishStepDocStringNotClosed):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_step_doc_string_should_preverse_comments(parser):
    """The parser should preserve gherkin comments in doc strings"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                # some comment
                My Doc String
                # another comment
                    foo
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == (
        "# some comment\n" "My Doc String\n" "# another comment\n" "    foo\n"
    )


def test_parse_step_empty_docstring(parser):
    """The parser should be able to parse an empty doc string"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                \"\"\"
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == ""


def test_parse_step_with_single_data_table_row(parser):
    """The parser should parse a Step with a single data table row"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].data_table == [["foo", "bar"]]


def test_parse_step_with_multiple_data_table_rows(parser):
    """The parser should parse a Step with multiple data table rows"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                | meh | rar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].data_table == [
        ["foo", "bar"],
        ["meh", "rar"],
    ]


def test_parse_step_with_single_data_table_row_with_consecutive_step(parser):
    """The parser should parse a Step with a single data table row followed by another Step"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                When there is a Step
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 2
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].data_table == [["foo", "bar"]]
    assert ast.rules[0].scenarios[0].steps[1].keyword == "When"
    assert ast.rules[0].scenarios[0].steps[1].text == "there is a Step"
    assert ast.rules[0].scenarios[0].steps[1].data_table is None


def test_parse_step_with_data_table_and_escaped_vbars(parser):
    """The parser should parse a Step with a data table containing escaped VBARs"""
    # given
    feature_file = r"""
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo\|bla | bar |
                | foo | bar\|bla |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].data_table == [
        ["foo|bla", "bar"],
        ["foo", "bar|bla"],
    ]


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                | meh | rar
                | bla | mop |
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                | meh
    """,
    ],
    ids=[
        "no trailing VBAR in first row",
        "no trailing VBAR in middle row",
        "no trailing VBAR in last row",
    ],
)
def test_parse_fail_for_missing_closing_vbar_in_step_data_table(parser, feature_file):
    """The parser should fail when encountering missing VBAR in Step Data Table"""
    # then
    with pytest.raises(RadishStepDataTableMissingClosingVBar):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo |
                | bar | meh |
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                | meh |
                | bla | mop |
    """,
        """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                | foo | bar |
                | meh | bla |
                | mop |
    """,
    ],
    ids=[
        "cell count mismatch in first row",
        "cell count mismatch in middle row",
        "cell count mismatch in last row",
    ],
)
def test_parse_fail_for_inconsistent_cell_count_in_step_data_table(
    parser, feature_file
):
    """The parser should fail when encountering inconsistent cell count in Step Data Table"""
    # then
    with pytest.raises(RadishStepDataTableInconsistentCellCount):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_step_doc_string_and_step_data_table(parser):
    """The parser should parse a Step with a Doc String and a Data Table"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario: My Scenario
                Given there is a setup
                \"\"\"
                My Doc String
                \"\"\"
                | foo | bar |
                | meh | bla |
                When there is a Step
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].steps) == 2
    assert ast.rules[0].scenarios[0].steps[0].keyword == "Given"
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a setup"
    assert ast.rules[0].scenarios[0].steps[0].doc_string == "My Doc String\n"
    assert ast.rules[0].scenarios[0].steps[0].data_table == [
        ["foo", "bar"],
        ["meh", "bla"],
    ]


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop

            Iterations: 2
    """,
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
            Iterations: 2
    """,
    ],
    ids=[
        "with blank line between Scenario Loop and Iterations",
        "without blank line between Scenario Loop and Iterations",
    ],
)
def test_parse_single_empty_scenario_loop(parser, feature_file):
    """The parser should parse a single Scenario Loop"""
    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioLoop)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Loop"
    assert ast.rules[0].scenarios[0].iterations == 2


def test_parse_example_loop_synonym_for_scenario_loop(parser):
    """The parser should recognize Example Loop as a synonym for Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            Example Loop: My Scenario Loop
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioLoop)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Loop"


def test_parse_single_scenario_loop_outside_rule(parser):
    """The parser should parse a single Scenario Loop outside any Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step
                When there is a Step
                Then there is a Step

            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Loop"
    assert ast.rules[0].scenarios[0].iterations == 2
    assert len(ast.rules[0].scenarios[0].steps) == 3


def test_parse_single_scenario_loop_inside_rule(parser):
    """The parser should parse a single Scenario Loop inside a Rule"""
    # given
    feature_file = """
        Feature: My Feature

            Rule: My Rule
                Scenario Loop: My Scenario Loop
                    Given there is a Step
                    When there is a Step
                    Then there is a Step

                Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My Rule"
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Loop"
    assert ast.rules[0].scenarios[0].iterations == 2
    assert len(ast.rules[0].scenarios[0].steps) == 3


def test_parse_fail_scenario_loop_without_iterations(parser):
    """The parser should fail if a Scenario Loop has no Iterations block"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step
    """

    # then
    with pytest.raises(RadishScenarioLoopMissingIterations):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step

            Iterations:
    """,
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step

            Iterations: str
    """,
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step

            Iterations: 0.5
    """,
        """
        Feature: My Feature

            Scenario Loop: My Scenario Loop
                Given there is a Step

            Iterations: 1 2
    """,
    ],
    ids=[
        "no Iterations value",
        "string as Iterations value",
        "float as Iterations value",
        "Iterations value with spaces",
    ],
)
def test_parse_fail_scenario_loop_without_iterations_value(parser, feature_file):
    """The parser should fail if a Scenario Loop has no value for the Iterations"""
    # then
    with pytest.raises(RadishScenarioLoopInvalidIterationsValue):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_tag_from_a_scenario_loop(parser):
    """The parser should parse a Tag from a Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            Scenario Loop: My Scenario
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 1
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"


def test_parse_tags_on_multiple_lines_from_a_scenario_loop(parser):
    """The parser should parse multiple Tags on multiple lines from a Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            @tag-b
            Scenario Loop: My Scenario
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 2
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"
    assert ast.rules[0].scenarios[0].tags[1].name == "tag-b"


def test_parser_build_examples_for_scenario_loop(parser):
    """The parser should build Examples for every Iteration of a Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Loop: My Scenario
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].examples) == 2


def test_parser_add_iteration_id_to_scenario_loop_example(parser):
    """
    The parser should build Examples for every Iteration of a
    Scenario Loop and add the Iteration Id to it's short description"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Loop: My Scenario
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert (
        ast.rules[0].scenarios[0].examples[0].short_description
        == "My Scenario [Iteration: 1]"
    )
    assert (
        ast.rules[0].scenarios[0].examples[1].short_description
        == "My Scenario [Iteration: 2]"
    )


def test_parser_built_examples_for_scenario_loop_have_copied_steps(parser):
    """The parser should copy the Steps for the built Examples of a Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Loop: My Scenario
                Given there is a Step

            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].examples) == 2
    assert len(ast.rules[0].scenarios[0].steps) == 1
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a Step"
    assert (
        ast.rules[0].scenarios[0].examples[0].steps[0]
        is not ast.rules[0].scenarios[0].steps[0]
    )
    assert (
        ast.rules[0].scenarios[0].examples[1].steps[0]
        is not ast.rules[0].scenarios[0].steps[0]
    )


def test_parser_assign_background_to_scenario_loop(parser):
    """The parser should assign the Background to the Scenario Loop"""
    # given
    feature_file = """
        Feature: My Feature

            Background:

            Scenario Loop: My Scenario
            Iterations: 2
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.background is not None
    assert ast.rules[0].scenarios[0].background is not None
    assert ast.rules[0].scenarios[0].examples[0].background is not None
    assert ast.rules[0].scenarios[0].examples[1].background is not None
    assert (
        ast.rules[0].scenarios[0].examples[0].background
        is not ast.rules[0].scenarios[0].examples[1].background
    )


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline

            Examples:
                | hdr |
                | foo |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
            Examples:
                | hdr |
                | foo |
    """,
    ],
    ids=[
        "with blank line between Scenario Outline and Examples",
        "without blank line between Scenario Outline and Examples",
    ],
)
def test_parse_single_empty_scenario_outline(parser, feature_file):
    """The parser should parse a single Scenario Outline"""
    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioOutline)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Outline"
    assert len(ast.rules[0].scenarios[0].examples) == 1


def test_parse_example_outline_synonym_for_scenario_outline(parser):
    """The parser should recognize Example Outline as a synonym for Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            Example Outline: My Scenario Outline
            Examples:
                | hdr |
                | foo |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioOutline)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Outline"


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
            Examples:
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
            Examples:
                | hdr |
    """,
    ],
    ids=[
        "no Examples block given without preceeding Steps",
        "no Examples block given with preceeding Steps",
        "no Examples table given",
        "only Examples Header given",
    ],
)
def test_parse_fail_no_examples_in_scenario_outline(parser, feature_file):
    """The parser should fail to parse if no Examples is given for a Scenario Outline"""
    # then
    with pytest.raises(RadishScenarioOutlineWithoutExamples):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                | foo
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                | foo |
                | bar
                | meh |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                | foo |
                | bar
    """,
    ],
    ids=[
        "no trailing VBAR in header",
        "no trailing VBAR in first row",
        "no trailing VBAR in middle row",
        "no trailing VBAR in last row",
    ],
)
def test_parse_fail_for_missing_closing_vbar_in_scenario_outline_examples(
    parser, feature_file
):
    """The parser should fail when encountering missing closing VBAR in Scenario Outline Examples"""
    # then
    with pytest.raises(RadishScenarioOutlineExamplesMissingClosingVBar):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        pytest.param(
            """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                hdr |
    """,
            marks=pytest.mark.xfail,
        ),
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                  foo |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                | foo |
                  bar |
                | meh |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr |
                | foo |
                  bar |
    """,
    ],
    ids=[
        "no opening VBAR in header",
        "no opening VBAR in first row",
        "no opening VBAR in middle row",
        "no opening VBAR in last row",
    ],
)
def test_parse_fail_for_missing_opening_vbar_in_scenario_outline_examples(
    parser, feature_file
):
    """The parser should fail when encountering missing opening VBAR in Scenario Outline Examples"""
    # then
    with pytest.raises(RadishScenarioOutlineExamplesMissingOpeningVBar):
        # when
        parser.parse_contents(None, feature_file)


@pytest.mark.parametrize(
    "feature_file",
    [
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr1 | hdr1 |
                | foo  |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr1 | hdr1 |
                | foo  |
                | bar  | meh  |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr1 | hdr1 |
                | foo  | bar  |
                | bar  |
                | meh  | bla  |
    """,
        """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a Step

            Examples:
                | hdr1 | hdr1 |
                | foo  | bar  |
                | bar  | meh  |
                | bla  |
    """,
    ],
    ids=[
        "cell count mismatch with header",
        "cell count mismatch in first row",
        "cell count mismatch in middle row",
        "cell count mismatch in last row",
    ],
)
def test_parse_fail_for_inconsistent_cell_count_in_scenario_outline_examples(
    parser, feature_file
):
    """
    The parser should fail when encountering inconsistent cell count in Scenario Outline Examples
    """
    # then
    with pytest.raises(RadishScenarioOutlineExamplesInconsistentCellCount):
        # when
        parser.parse_contents(None, feature_file)


def test_parse_single_scenario_outline_outside_rule(parser):
    """The parser should parse a single Scenario Outline outside any Rule"""
    feature_file = """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioOutline)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Outline"
    assert len(ast.rules[0].scenarios[0].examples) == 1


def test_parse_single_scenario_outline_inside_rule(parser):
    """The parser should parse a single Scenario Outline inside a Rule"""
    feature_file = """
        Feature: My Feature

            Rule: My Rule
                Scenario Outline: My Scenario Outline
                    Given there is a <hdr> Step
                    When there is a <hdr> Step
                    Then there is a <hdr> Step

                Examples:
                    | hdr |
                    | foo |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].short_description == "My Rule"
    assert len(ast.rules[0].scenarios) == 1
    assert isinstance(ast.rules[0].scenarios[0], ScenarioOutline)
    assert ast.rules[0].scenarios[0].short_description == "My Scenario Outline"
    assert len(ast.rules[0].scenarios[0].examples) == 1


def test_parse_tag_from_a_scenario_outline(parser):
    """The parser should parse a Tag from a Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
                | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 1
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"


def test_parse_tags_on_multiple_lines_from_a_scenario_outline(parser):
    """The parser should parse multiple Tags on multiple lines from a Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            @tag-a
            @tag-b
            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
                | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].tags) == 2
    assert ast.rules[0].scenarios[0].tags[0].name == "tag-a"
    assert ast.rules[0].scenarios[0].tags[1].name == "tag-b"


def test_parser_build_examples_for_scenario_outline(parser):
    """The parser should build Examples for every Iteration of a Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
                | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].examples) == 2


def test_parser_add_example_data_to_short_description_for_scenario_outline_examples(
    parser,
):
    """
    The parser should build Examples for every Iteration of a Scenario Outline
    and add the Example Data to it's short description
    """
    # given
    feature_file = """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr1 | hdr2 |
                | foo  | meh  |
                | bar  | bla  |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    examples = ast.rules[0].scenarios[0].examples
    assert (
        examples[0].short_description == "My Scenario Outline [hdr1: foo, hdr2: meh]"
        # Python 3.5 has no dict ordering
        or examples[0].short_description == "My Scenario Outline [hdr2: meh, hdr1: foo]"
    )

    assert (
        examples[1].short_description == "My Scenario Outline [hdr1: bar, hdr2: bla]"
        # Python 3.5 has no dict ordering
        or examples[1].short_description == "My Scenario Outline [hdr2: bla, hdr1: bar]"
    )


def test_parser_built_examples_for_scenario_outline_have_copied_steps(parser):
    """The parser should copy the Steps for the built Examples of a Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
                | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert len(ast.rules[0].scenarios[0].examples) == 2
    assert len(ast.rules[0].scenarios[0].steps) == 3
    assert ast.rules[0].scenarios[0].steps[0].text == "there is a <hdr> Step"
    assert (
        ast.rules[0].scenarios[0].examples[0].steps[0]
        is not ast.rules[0].scenarios[0].steps[0]
    )
    assert (
        ast.rules[0].scenarios[0].examples[1].steps[0]
        is not ast.rules[0].scenarios[0].steps[0]
    )


def test_parser_assign_background_to_scenario_outline(parser):
    """The parser should assign the Background to the Scenario Outline"""
    # given
    feature_file = """
        Feature: My Feature

            Background:

            Scenario Outline: My Scenario Outline
                Given there is a <hdr> Step
                When there is a <hdr> Step
                Then there is a <hdr> Step

            Examples:
                | hdr |
                | foo |
                | bar |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.background is not None
    assert ast.rules[0].scenarios[0].background is not None
    assert ast.rules[0].scenarios[0].examples[0].background is not None
    assert ast.rules[0].scenarios[0].examples[1].background is not None
    assert (
        ast.rules[0].scenarios[0].examples[0].background
        is not ast.rules[0].scenarios[0].examples[1].background
    )


def test_parser_replace_examples_parameter_in_scenario_outline_examples(parser):
    """The parser should replcae the Examples parameter in the built Scenario Outline Examples"""
    # given
    feature_file = """
        Feature: My Feature

            Scenario Outline: My Scenario Outline
                Given there is a <hdr1> Step
                When there is a <hdr2> Step
                Then there is a <hdr1>-<hdr2> Step

            Examples:
                | hdr1 | hdr2 |
                | foo  | bar  |
                | meh  | bla  |
    """

    # when
    ast = parser.parse_contents(None, feature_file)

    # then
    assert ast.rules[0].scenarios[0].examples[0].steps[0].text == "there is a foo Step"
    assert ast.rules[0].scenarios[0].examples[0].steps[1].text == "there is a bar Step"
    assert (
        ast.rules[0].scenarios[0].examples[0].steps[2].text == "there is a foo-bar Step"
    )

    assert ast.rules[0].scenarios[0].examples[1].steps[0].text == "there is a meh Step"
    assert ast.rules[0].scenarios[0].examples[1].steps[1].text == "there is a bla Step"
    assert (
        ast.rules[0].scenarios[0].examples[1].steps[2].text == "there is a meh-bla Step"
    )


def test_parser_precondition_tag_on_scenario(parser):
    """The parser should recognize a precondition Tag on a Scenario"""
    # given
    feature_file = """
        Feature: My Feature

            @precondition(some.feature: My Base Scenario)
            Scenario: My Scenario with a Precondition
                Given there is a Step
    """

    # when
    ast = parser.parse_contents(Path(__file__), feature_file)

    # then
    scenario = ast.rules[0].scenarios[0]
    precondition_tag = scenario.tags[0]

    assert isinstance(precondition_tag, PreconditionTag)
    assert precondition_tag.name == "precondition(some.feature: My Base Scenario)"
    assert precondition_tag.feature_filename == "some.feature"
    assert precondition_tag.scenario_short_description == "My Base Scenario"


def test_parser_constant_tag_on_feature(parser):
    """The parser should recognize a constant Tag on a Feature"""
    # given
    feature_file = """
        @constant(foo: bar)
        Feature: My Feature
    """

    # when
    ast = parser.parse_contents(Path(__file__), feature_file)

    # then
    assert isinstance(ast.tags[0], ConstantTag)


def test_parser_should_raise_error_for_unknown_language(parser):
    """The parser should raise an error for an unknown language"""
    # given
    feature_file = """
        # language: FOO
        Feature: My Feature
    """
    # then
    with pytest.raises(
        RadishLanguageNotFound, match="The language FOO is currently not supported"
    ):
        # when
        parser.parse_contents(Path(__file__), feature_file)


def test_parser_should_recognize_the_language(parser):
    """The parser should recognize the specified language in the Feature File"""
    # given
    feature_file = """
        # language: de
        Funktionalität: Meine Funktionalität

            Szenario: Mein Szenario
                Gegeben ist ein Schritt
    """

    # when
    ast = parser.parse_contents(Path(__file__), feature_file)

    # then
    assert len(ast.rules[0].scenarios) == 1
    assert ast.keyword == "Funktionalität"



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from datetime import timedelta

from freezegun import freeze_time

from radish.models.timed import Timed


def test_timed_model_should_be_created_without_a_start_and_end_time():
    """A Timed model should be created without a start- and end time"""
    # when
    model = Timed()

    # then
    assert model.starttime is None
    assert model.endtime is None


@freeze_time("1994-02-21")
def test_timed_model_should_record_current_time_when_recording_started():
    """A Timed model should record the current time when the recording is started"""
    # given
    model = Timed()

    # when
    model.record_starttime()

    # then
    model.starttime == 761788800


@freeze_time("1994-02-21")
def test_timed_model_should_record_current_time_when_recording_ended():
    """A Timed model should record the current time when the recording is ended"""
    # given
    model = Timed()

    # when
    model.record_endtime()

    # then
    model.starttime == 761788800


def test_timed_model_zero_duration_when_recording_not_started():
    """A Timed model should calculated a duration of ZERO when the recording was not started"""
    # given
    model = Timed()

    # when
    duration = model.duration()

    # then
    assert duration == timedelta(0)


@freeze_time("1994-02-22")
def test_timed_model_should_give_the_current_duration_when_only_starttime_recorded():
    """
    A Timed model should calulated the duration until the
    current time when only the starttime is recorded
    """
    # given
    model = Timed()
    model.starttime = 761788800  # 1994-02-21

    # when
    duration = model.duration()

    # then
    assert duration == timedelta(days=1)


@freeze_time("1994-02-22")
def test_timed_model_should_calculate_the_duration_between_recorded_start_and_end_time():
    """
    A Timed model should calulated the duration between
    a recorded start- and end time.
    """
    # given
    model = Timed()
    model.starttime = 761788800  # 1994-02-21
    model.endtime = model.starttime + 60  # plus a minute

    # when
    duration = model.duration()

    # then
    assert duration == timedelta(minutes=1)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import importlib.util
from pathlib import Path
from typing import List

from radish.errors import RadishError


def load_modules(locations: List[Path]) -> List[Path]:
    """Load all Python modules in the given locations

    All given locations must already be expanded regarding
    * Environment Variables
    * User Home Directory
    """
    loaded_modules = []
    for location in locations:
        if not location.exists():
            raise RadishError(
                "Location '{0}' to load modules does not exist".format(location)
            )

        for module_path in location.glob("**/*.py"):
            load_module(module_path)
            loaded_modules.append(module_path)

    return loaded_modules


def load_module(path: Path) -> None:
    """Load the given module into the Python runtime"""
    module_name = path.stem
    try:
        spec = importlib.util.spec_from_file_location(module_name, str(path))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except Exception as exc:
        raise ImportError(
            "Unable to import module '{0}' from '{1}': {2}".format(
                module_name, path, exc
            )
        ) from exc



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import sys
import traceback


class StepFailureReport:
    """Represents a Failure Report for a single Step"""

    def __init__(self, exception):
        self.exception = exception
        self.reason = str(exception)
        self.traceback = str(traceback.format_exc())
        self.name = exception.__class__.__name__
        traceback_info = traceback.extract_tb(sys.exc_info()[2])[-1]
        self.filename = traceback_info[0]
        self.line = int(traceback_info[1])



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from collections import namedtuple
from pathlib import Path

import click
import colorful as cf

import radish.loader as loader
from radish.__main__ import expand_basedirs
from radish.config import Config
from radish.logger import enable_radish_debug_mode_click_hook, logger
from radish.stepregistry import registry as step_registry

from radish.step_testing.matcher import run_matcher_tests


#: Holds a type for the coverage configuration
CoverageConfig = namedtuple(
    "CoverageConfig", ["show_missing", "show_missing_templates"]
)


def expand_matcher_configs(ctx, param, matcher_configs):
    """Expand the given matcher configuration files

    Expanding directories recursively for YAML files.
    """
    expanded_matcher_configs = []
    for matcher_config_file_location in (Path(f) for f in matcher_configs):
        if matcher_config_file_location.is_dir():
            expanded_matcher_configs.extend(
                list(matcher_config_file_location.glob("**/*.yml"))
            )
        else:
            expanded_matcher_configs.append(matcher_config_file_location)

    return expanded_matcher_configs


@click.command()
@click.version_option()
@click.help_option("--help", "-h")
@click.option(
    "--debug",
    "-d",
    "enable_debug_mode",
    is_flag=True,
    help="Enable debug mode for radish-test itself",
    callback=enable_radish_debug_mode_click_hook,
)
@click.option(
    "--basedir",
    "-b",
    "basedirs",
    multiple=True,
    default=("radish",),
    type=str,
    callback=expand_basedirs,
    help=(
        "Specify the location of the Step Implementations. "
        "One '-b' can contain multiple locations, split by a colon (:) [UNIX] "
        "or semicolon (;) [Windows]"
    ),
)
@click.option(
    "--no-ansi",
    "no_ansi",
    is_flag=True,
    help="Turn off all ANSI sequences (colors, line rewrites).",
)
@click.option(
    "--show-missing",
    "-m",
    "show_missing",
    is_flag=True,
    help="Show all Step Implementations which are not yet covered by a test",
)
@click.option(
    "--show-missing-templates",
    "show_missing_templates",
    is_flag=True,
    help="Print templates for all missing Step Implementations. Implies --show-missing.",
)
@click.argument(
    "matcher-configs",
    nargs=-1,
    type=click.Path(exists=True),
    callback=expand_matcher_configs,
)
def cli(**kwargs):
    """radish - The root from red to green. BDD tooling for Python.

    radish-test can be used to perform tests for the Steps
    implemented in a radish base directory.

    Use the `MATCHER_CONFIGS` to pass configuration
    files containing the Step matchers.
    """
    config = Config(kwargs)

    # turn of ANSI colors if requested
    if config.no_ansi:
        cf.disable()

    logger.debug("Basedirs: %s", ", ".join(str(d) for d in config.basedirs))
    logger.debug("Loading all modules from the basedirs")
    loaded_modules = loader.load_modules(config.basedirs)
    logger.debug(
        "Loaded %d modules from the basedirs: %s",
        len(loaded_modules),
        ", ".join(str(m) for m in loaded_modules),
    )

    logger.debug(
        "Matcher configs: %s", ", ".join(str(m) for m in config.matcher_configs)
    )

    coverage_config = CoverageConfig(config.show_missing, config.show_missing_templates)

    run_matcher_tests(config.matcher_configs, coverage_config, step_registry)


if __name__ == "__main__":
    cli()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.models import Rule, State


def test_rule_should_forward_set_feature_to_its_scenarios(mocker):
    """A Rule should forward a set Feature to its Scenarios"""
    # given
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    rule = Rule("Rule", "My Rule", None, None, [first_scenario, second_scenario])
    feature_mock = mocker.MagicMock(name="Feature")

    # when
    rule.set_feature(feature_mock)

    # then
    assert rule.feature is feature_mock
    first_scenario.set_feature.assert_called_once_with(feature_mock)
    second_scenario.set_feature.assert_called_once_with(feature_mock)


def test_rule_should_forward_set_background_to_its_scenarios(mocker):
    """A Rule should forward a set Background to its Scenarios"""
    # given
    first_scenario = mocker.MagicMock(name="First Scenario")
    second_scenario = mocker.MagicMock(name="Second Scenario")
    rule = Rule("Rule", "My Rule", None, None, [first_scenario, second_scenario])
    background_mock = mocker.MagicMock(name="Background")

    # when
    rule.set_background(background_mock)

    # then
    first_scenario.set_background.assert_called_once_with(background_mock)
    second_scenario.set_background.assert_called_once_with(background_mock)


@pytest.mark.parametrize(
    "given_scenario_states, expected_state",
    [
        ([State.PASSED, State.PASSED, State.PASSED], State.PASSED),
        ([State.PASSED, State.UNTESTED, State.PASSED], State.UNTESTED),
        ([State.PASSED, State.SKIPPED, State.UNTESTED], State.SKIPPED),
        ([State.PASSED, State.PENDING, State.SKIPPED], State.PENDING),
        ([State.PASSED, State.PENDING, State.FAILED], State.FAILED),
        ([State.PASSED, State.RUNNING, State.FAILED], State.RUNNING),
    ],
    ids=[
        "[State.PASSED, State.PASSED, State.PASSED] -> State.PASSED",
        "[State.PASSED, State.UNTESTED, State.PASSED] -> State.UNTESTED",
        "[State.PASSED, State.SKIPPED, State.UNTESTED] -> State.SKIPPED",
        "[State.PASSED, State.PENDING, State.SKIPPED] -> State.PENDING",
        "[State.PASSED, State.PENDING, State.FAILED] -> State.FAILED",
        "[State.PASSED, State.RUNNING, State.FAILED] -> State.RUNNING",
    ],
)
def test_rule_should_return_correct_state_according_to_its_scenario_states(
    given_scenario_states, expected_state, mocker
):
    """A Rule should return the correct State according to its Scenario States"""
    # given
    rule = Rule(
        "Rule",
        "My Rule",
        None,
        None,
        [mocker.MagicMock(state=s) for s in given_scenario_states],
    )

    # when
    actual_state = rule.state

    # then
    assert actual_state == expected_state


@pytest.mark.parametrize(
    "scenarios_need_to_run, expected_has_to_run",
    [
        ([False, False, False], False),
        ([False, True, False], True),
        ([False, True, True], True),
    ],
    ids=[
        "no Scenario needs to run",
        " a Scenario needs to run",
        "multiple Scenarios need to run",
    ],
)
def test_rule_should_run_if_one_of_its_scenario_has_to_run(
    scenarios_need_to_run, expected_has_to_run, mocker
):
    """A Rule should run if one of its Scenarios has to run"""
    scenarios = []
    for has_to_run in scenarios_need_to_run:
        scenario_mock = mocker.MagicMock()
        scenario_mock.has_to_run.return_value = has_to_run
        scenarios.append(scenario_mock)

    # given
    rule = Rule("Rule", "My Rule", None, None, scenarios)

    # when
    actual_has_to_run = rule.has_to_run(None, None)

    # then
    assert actual_has_to_run == expected_has_to_run



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

from radish.terrain import world
from radish.config import Config


@pytest.fixture(name="default_config", scope="function")
def setup_default_config():
    """Setup a default radish config

    This config can be used by a components under test for most Test Cases.
    """
    return Config(
        {
            "wip_mode": False,
            "dry_run_mode": False,
            "debug_steps_mode": False,
            "tags": None,
            "tag_expression": None,
            "scenario_ids": [],
            "early_exit": False,
            "shuffle_scenarios": False,
            "no_ansi": False,
        }
    )


@pytest.fixture(name="world_default_config", scope="function")
def setup_world_default_config(default_config):
    """Setup a default radish config in the terrain world object"""
    if hasattr(world, "config"):
        orig_config = world.config
    else:
        orig_config = None

    world.config = default_config
    yield world.config

    if orig_config is not None:
        world.config = orig_config
    else:
        delattr(world, "config")



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import click

from radish.utils import get_debugger
from radish.extensionregistry import extension
from radish.hookregistry import after
from radish.models.state import State


@extension
class FailureDebugger:
    """
    Extension to start a debugger upon a Step failure
    """

    OPTIONS = [
        click.Option(
            param_decls=("--debug-after-failure", "debug_after_failure"),
            is_flag=True,
            help="Start a Python debugger when a Step failed",
        )
    ]

    @classmethod
    def load(cls, config):
        if config.debug_after_failure:
            return cls()
        else:
            return None

    def __init__(self):
        after.each_step()(failure_debugger)


def failure_debugger(self, step):
    """Start the Python debugger after a Step failure"""
    if step.state is not State.FAILED:
        return

    pdb = get_debugger()
    pdb.set_trace()



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.constant_tag import ConstantTag
from radish.models.context import Context
from radish.models.state import State
from radish.models.timed import Timed


class Feature(Timed):
    """"Represents a single instance of a Gherkin Feature"""

    def __init__(
        self,
        feature_id: int,
        keyword: str,
        short_description: str,
        description,
        tags,
        path: str,
        line: int,
        background,
        rules,
        language_spec,
    ) -> None:
        super().__init__()
        self.id = feature_id
        self.keyword = keyword
        self.short_description = short_description
        self.description = description
        self.tags = tags
        self.path = path
        self.line = line
        self.background = background
        self.rules = rules
        self.language_spec = language_spec

        #: Holds the ``Context`` object.
        self.context = Context()

    def __repr__(self) -> str:
        return "<Feature: {id} '{short_description}' with {rules} Rules @ {path}:{line}>".format(
            id=self.id,
            short_description=self.short_description,
            rules=len(self.rules),
            path=self.path,
            line=self.line,
        )  # pragma: no cover

    @property
    def state(self):
        """Read-only property to get the State for this Feature"""
        return State.report_state(r.state for r in self.rules)

    @property
    def constants(self):
        """Get the Constants of this Feature

        The Constants are lazy-evaluated from the Tags.
        """
        return {t.key: t.value for t in self.tags if isinstance(t, ConstantTag)}

    def get_all_tags(self):
        """Return all Tags for this Feature"""
        return self.tags

    def has_to_run(self, tag_expression, scenario_ids):
        """Evaluate if this Feature has to run or not

        A Feature has to run of any of it's Rules has to be run.
        """
        return any(r.has_to_run(tag_expression, scenario_ids) for r in self.rules)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.terrain import world


def test_pick_function_into_world_object():
    """
    The world.pick() decroator should be able to assign arbitrary functions to the world object
    """
    # when
    @world.pick
    def some_arbitrary_helper_func(answer):
        return answer == 42

    # then
    assert world.some_arbitrary_helper_func is some_arbitrary_helper_func



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.models.tag import Tag


class ConstantTag(Tag):
    """Represents a single radish Constant Tag"""

    def __init__(self, key: str, value: str, path: str, line: int) -> None:
        super().__init__("constant({}: {})".format(key, value), path, line)
        self.key = key
        self.value = value

    def __repr__(self) -> str:
        return "<Constant: {name} @ {path}:{line}>".format(
            name=self.name, path=self.path, line=self.line
        )  # pragma: no cover



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from radish.errors import RadishError


class RadishSyntaxError(RadishError, SyntaxError):
    def __str__(self):
        context, line, column = self.args
        return "{} at line {}, column {}.\n\n{}".format(
            self.label, line, column, context
        )


class RadishMissingFeatureShortDescription(RadishSyntaxError):
    label = "Feature is missing a short description"
    examples = ["Feature:"]


class RadishMissingRuleShortDescription(RadishSyntaxError):
    label = "Rule is missing a short description"
    examples = [
        """
        Feature: some feature

            Rule:
        """
    ]


class RadishMissingScenarioShortDescription(RadishSyntaxError):
    label = "Scenario is missing a short description"
    examples = [
        """
        Feature: some feature

            Scenario:
        """
    ]


class RadishMultipleBackgrounds(RadishSyntaxError):
    label = "Maximum one Background is allowed per Feature"
    examples = [
        """
        Feature: some feature

            Backround: My first Background
            Background: My second Background
        """
    ]


class RadishMisplacedBackground(RadishSyntaxError):
    label = "Background must be the first block in a Feature"
    examples = [
        """
        Feature: some feature
                Test

            Scenario: some scenario
                When I do

            Background: some background
                When I do
        """
    ]


class RadishStepDoesNotStartWithKeyword(RadishSyntaxError):
    label = "Steps must start with a step keyword: Given, When, Then, And, But"
    examples = [
        """
        Feature: some feature
            Test

            Scenario: some scenario
                I do
        """
    ]


class RadishStepDocStringNotClosed(RadishSyntaxError):
    label = (
        'A Step Doc String must be closed with three-consecutive double quotes: """' ""
    )
    examples = [
        """
        Feature: some feature
            Test

            Scenario: some scenario
                Given there is a Step
                \"\"\"
                My doc string
                When there is a Step
        """,
        """
        Feature: some feature
            Test

            Scenario: some scenario
                Given there is a Step
                \"\"\"
                My doc string
                "
                When there is a Step
        """,
        """
        Feature: some feature
            Test

            Scenario: some scenario
                Given there is a Step
                \"\"\"
                My doc string
                ""
                When there is a Step
        """,
    ]


class RadishScenarioOutlineWithoutExamples(RadishSyntaxError):
    label = "Scenario Outlines must have an Examples block"
    examples = [
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                Given there is a Step
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do

            Examples:
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do

            Examples:
                | hdr |
        """,
    ]


class RadishStepDataTableMissingClosingVBar(RadishSyntaxError):
    label = "Missing closing VBAR in Step Data Table"
    examples = [
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do
                | foo
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do
                | foo | bar
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do
                | foo | bar |
                | meh
        """,
    ]


class RadishStepDataTableInconsistentCellCount(RadishSyntaxError):
    label = "Step Data Table rows must all have the same amount of cells"
    examples = [
        """
            Feature: My Feature

                Scenario: My Scenario
                    Given there is a setup
                    | foo |
                    | bar | meh |
        """,
        """
            Feature: My Feature

                Scenario: My Scenario
                    Given there is a setup
                    | foo | bar |
                    | meh |
                    | bla | mop |
        """,
        """
            Feature: My Feature

                Scenario: My Scenario
                    Given there is a setup
                    | foo | bar |
                    | meh | bla |
                    | mop |
        """,
    ]


class RadishScenarioLoopMissingIterations(RadishSyntaxError):
    label = "The Scenario Loop is missing the Iterations block"
    examples = [
        """
            Feature: My Feature

                Scenario Loop: My Scenario
                    Given there is a setup
        """,
        """
            Feature: My Feature

                Scenario Loop: My Scenario
        """,
    ]


class RadishScenarioLoopInvalidIterationsValue(RadishSyntaxError):
    label = "The Scenario Loop is an invalid Iterations value"
    examples = [
        """
            Feature: My Feature

                Scenario Loop: My Scenario Loop
                    Given there is a Step

                Iterations:
        """,
        """
            Feature: My Feature

                Scenario Loop: My Scenario Loop
                    Given there is a Step

                Iterations: str
        """,
        """
            Feature: My Feature

                Scenario Loop: My Scenario Loop
                    Given there is a Step

                Iterations: 0.5
        """,
        """
            Feature: My Feature

                Scenario Loop: My Scenario Loop
                    Given there is a Step

                Iterations: 1 2
        """,
    ]


class RadishScenarioOutlineExamplesMissingClosingVBar(RadishSyntaxError):
    label = "Missing closing VBAR in Scenario Outline Examples"
    examples = [
        """
            Feature: My Feature

                Scenario Outline: My Scenario Outline
                    Given there is a Step

                Examples:
                    |
        """,
        """
            Feature: My Feature

                Scenario Outline: My Scenario Outline
                    Given there is a Step

                Examples:
                    | hdr
        """,
        """
            Feature: My Feature

                Scenario Outline: My Scenario Outline
                    Given there is a Step

                Examples:
                    | hdr |
                    | foo
        """,
        """
            Feature: My Feature

                Scenario Outline: My Scenario Outline
                    Given there is a Step

                Examples:
                    | hdr |
                    | foo |
                    | bar
                    | meh |
        """,
        """
            Feature: My Feature

                Scenario Outline: My Scenario Outline
                    Given there is a Step

                Examples:
                    | hdr |
                    | foo |
                    | bar
        """,
    ]


class RadishScenarioOutlineExamplesMissingOpeningVBar(RadishSyntaxError):
    label = "Missing opening VBAR in Scenario Outline Examples"
    examples = [
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do

            Examples:
                foo
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do

            Examples:
                foo |
        """,
        """
        Feature: some feature
            Test

            Scenario Outline: some scenario
                When I do

            Examples:
                | hdr |
                  foo |
        """,
    ]


class RadishScenarioOutlineExamplesInconsistentCellCount(RadishSyntaxError):
    label = "Scenario Outline Examples rows must all have the same amount of cells"
    examples = []


class RadishFirstStepMustUseFirstLevelKeyword(RadishSyntaxError):
    label = (
        "The first Step in a Scenario must use one of the first level Step keywords: "
        "Given, When or Then"
    )
    examples = []


class RadishPreconditionScenarioDoesNotExist(RadishSyntaxError):
    examples = []

    def __init__(
        self, featurefile_path, scenario_short_description, scenario_short_descriptions
    ):
        self.label = "No Scenario '{}' found in '{}' to use as a Precondition. Use one of: {}".format(  # noqa
            scenario_short_description,
            featurefile_path,
            ", ".join(scenario_short_descriptions),
        )


class RadishPreconditionScenarioRecursion(RadishSyntaxError):
    examples = []

    def __init__(self, scenario, precondition_scenario):
        self.label = "Detected a Precondition Recursion in '{}' caused by '{}'".format(
            scenario.short_description, precondition_scenario.short_description
        )

    def __str__(self):
        return self.label


class RadishLanguageNotFound(RadishSyntaxError):
    examples = []

    def __init__(self, language):
        self.label = "The language {} is currently not supported".format(language)

    def __str__(self):
        return self.label



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest

import radish.utils as utils


@pytest.mark.filterwarnings("ignore")
def test_getting_any_debugger():
    """When asking for a debugger it should return one

    It shouldn't matter if IPython is installed or not,
    just give me that debugger.
    """
    # when
    debugger = utils.get_debugger()

    # then
    assert callable(debugger.runcall)


def test_utils_should_locate_arbitrary_python_object():
    # when
    obj = utils.locate_python_object("str")

    # then
    assert obj == str


def test_converting_pos_args_into_kwargs():
    # given
    def func(_, arg1, arg2, kwarg1=1, kwargs2=2):
        pass

    pos_arg_values = ["arg1-value", "arg2-value"]

    # when
    kwargs = utils.get_func_pos_args_as_kwargs(func, pos_arg_values)

    # then
    assert kwargs == {"arg1": "arg1-value", "arg2": "arg2-value"}



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re
import textwrap

# NOTE(TF): we need the OrderedDict for Python 3.5 test reproducibility.
from collections import OrderedDict

import pytest

from radish.errors import RadishError, StepImplementationNotFoundError
from radish.models import Step
from radish.step_testing.matcher import (
    assert_step_arguments,
    assert_step_match,
    assert_step_not_match,
    parse_matcher_config,
    print_failure,
)
from radish.stepregistry import StepRegistry


def create_matcher_config_file(tmp_path, contents):
    """Helper function to create a matcher config file from a string"""
    tmp_matcher_path = tmp_path / "matcher_config.yml"
    with tmp_matcher_path.open("w", encoding="utf-8") as tmp_matcher_file:
        tmp_matcher_file.write(textwrap.dedent(contents))

    return tmp_matcher_path


def test_parse_config_should_fail_if_step_key_not_present(tmp_path):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - not-step: step text

        - not-step: step text
        """,
    )

    # THEN
    with pytest.raises(
        RadishError, match="All Matcher Config must start with the 'step' key"
    ):
        # WHEN
        parse_matcher_config(matcher_config_path)


def test_parse_config_should_fail_if_should_match_keys_not_present(tmp_path):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
        """,
    )

    # THEN
    with pytest.raises(
        RadishError,
        match=(
            "Each Matcher Config must either contain "
            "a 'should_match' or 'should_not_match' key"
        ),
    ):
        # WHEN
        parse_matcher_config(matcher_config_path)


def test_parse_config_should_fail_if_should_match_key_and_should_match_present(
    tmp_path,
):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
          should_match: step func
          should_not_match: step func
        """,
    )

    # THEN
    with pytest.raises(
        RadishError,
        match=(
            "Each Matcher Config must either contain "
            "a 'should_match' or 'should_not_match' key"
        ),
    ):
        # WHEN
        parse_matcher_config(matcher_config_path)


def test_parse_config_should_pass_if_should_match_key_not_and_should_not_match_present(
    tmp_path,
):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
          should_not_match: step func
        """,
    )

    # WHEN
    match_config = parse_matcher_config(matcher_config_path)

    # THEN
    assert match_config[0]["should_not_match"] == "step func"


def test_parse_config_should_pass_if_should_not_match_key_not_and_should_match_present(
    tmp_path,
):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
          should_match: step func
        """,
    )

    # WHEN
    match_config = parse_matcher_config(matcher_config_path)

    # THEN
    assert match_config[0]["should_match"] == "step func"


def test_parse_config_should_pass_if_with_arguments_is_specified(tmp_path):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
          should_match: step func
          with_arguments:
            - foo: bar
        """,
    )

    # WHEN
    match_config = parse_matcher_config(matcher_config_path)

    # THEN
    assert match_config[0]["with_arguments"] == [{"foo": "bar"}]


def test_parse_config_should_fail_for_invalid_keys(tmp_path):
    # GIVEN
    matcher_config_path = create_matcher_config_file(
        tmp_path,
        """
        - step: step text
          should_match: step func
          invalid_key: x
          another_invalid_key: y
        """,
    )

    # THEN
    with pytest.raises(
        RadishError,
        match=(
            "All Matcher Config only allow the keys: "
            "should_match, should_not_match, step, with_arguments. "
            "The following are not allowed: another_invalid_key, invalid_key."
        ),
    ):
        # WHEN
        parse_matcher_config(matcher_config_path)


def test_print_failure_should_output_only_red_cross_if_no_func_and_no_errors(capsys):
    # GIVEN
    expected_output = "✘ \n"

    # WHEN
    print_failure(None, [])
    actual_output = capsys.readouterr().out

    # THEN
    assert actual_output == expected_output


def test_print_failure_should_output_func_location(capsys):
    # GIVEN
    def func():
        pass

    expected_output = r"✘ \(at .*?test_step_testing.py:{}\)\n".format(
        func.__code__.co_firstlineno
    )

    # WHEN
    print_failure(func, [])
    actual_output = capsys.readouterr().out

    # THEN
    assert re.match(expected_output, actual_output)


def test_print_failure_should_output_specified_errors(capsys):
    # GIVEN
    errors = ["some error", "another error"]

    expected_output = "✘ \n  - some error\n  - another error\n"

    # WHEN
    print_failure(None, errors)
    actual_output = capsys.readouterr().out

    # THEN
    assert actual_output == expected_output


def test_assert_step_arguments_should_fail_if_expected_argument_is_not_present():
    # GIVEN
    actual_step_arguments = {"arg1": None, "arg2": None}
    expected_step_arguments = {"not-existing-arg": "value1"}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == (
        "Expected argument \"not-existing-arg\" is not in matched arguments ['arg1', 'arg2']"
    )


def test_assert_step_arguments_should_pass_if_arguments_match():
    # GIVEN
    actual_step_arguments = {"arg1": "value", "arg2": 2}
    expected_step_arguments = {"arg1": "value", "arg2": 2}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 0


def test_assert_step_arguments_should_fail_if_expected_argument_value_doesnt_match():
    # GIVEN
    actual_step_arguments = {"arg1": "value", "arg2": 2}
    expected_step_arguments = {"arg1": "not-correct-value", "arg2": 2}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == (
        'Expected argument "arg1" with value "not-correct-value" does not match value "value"'
    )


def test_assert_step_arguments_should_collect_all_argument_mismatched():
    # GIVEN
    actual_step_arguments = {"arg1": "value", "arg2": 2}
    expected_step_arguments = OrderedDict([("arg1", "not-correct-value"), ("arg2", -1)])

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 2
    assert actual_errors[0] == (
        'Expected argument "arg1" with value "not-correct-value" does not match value "value"'
    )
    assert actual_errors[1] == (
        'Expected argument "arg2" with value "-1" does not match value "2"'
    )


def test_assert_step_arguments_should_fail_if_argument_values_have_type_mismatch():
    # GIVEN
    actual_step_arguments = {"int-arg1": 42}
    expected_step_arguments = {"int-arg1": "42"}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == (
        'Expected argument "int-arg1" is of type "str" but should be "int"'
    )


def test_assert_step_arguments_should_correctly_compare_dicts():
    # GIVEN
    actual_step_arguments = {"dict-arg1": {"key1": 1, "key2": 2}}
    expected_step_arguments = {"dict-arg1": {"key1": 1, "key2": 2}}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 0


def test_assert_step_arguments_should_fail_if_dicts_dont_match():
    # GIVEN
    actual_step_arguments = {"dict-arg1": {"key1": 1, "key2": 2}}
    expected_step_arguments = {"dict-arg1": {"key1": 0, "key2": 0}}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1


def test_assert_step_arguments_should_use_explcitly_specified_type_and_value():
    # GIVEN
    actual_step_arguments = {"str-arg1": "value"}
    expected_step_arguments = {"str-arg1": {"type": "str", "value": "value"}}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 0


def test_assert_step_arguments_should_use_repr_if_specified_along_with_explcit_type_and_value():
    # GIVEN
    actual_step_arguments = {"str-arg1": "value"}
    expected_step_arguments = {
        "str-arg1": {"use_repr": True, "type": "str", "value": "'value'"}
    }

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 0


def test_assert_step_arguments_should_fail_if_specified_value_has_not_explictly_specified_type():
    # GIVEN
    actual_step_arguments = {"str-arg1": "value"}
    expected_step_arguments = {"str-arg1": {"type": "str", "value": 42}}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == (
        "Conflicting argument configuration: "
        'given value is actually of type "int" although it should match a value of type "str"'
    )


def test_assert_step_arguments_should_use_specified_cast_for_explicit_value():
    # GIVEN
    actual_step_arguments = {"int-arg1": 42}
    expected_step_arguments = {"int-arg1": {"type": "int", "value": "42", "cast": True}}

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 0


def test_assert_step_arguments_should_fail_if_type_is_not_known_for_cast():
    # GIVEN
    actual_step_arguments = {"int-arg1": 42}
    expected_step_arguments = {
        "int-arg1": {"type": "UnknownType", "value": "42", "cast": True}
    }

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == (
        'Cannot cast to type "UnknownType" because it is unknown'
    )


def test_assert_step_arguments_should_fail_if_unable_to_cast():
    # GIVEN
    actual_step_arguments = {"int-arg1": 42}
    expected_step_arguments = {
        "int-arg1": {"type": "int", "value": "not-an-int", "cast": True}
    }

    # WHEN
    actual_errors = assert_step_arguments(
        actual_step_arguments, expected_step_arguments
    )

    # THEN
    assert len(actual_errors) == 1
    assert actual_errors[0] == ('Failed to cast "not-an-int" to given type "int"')


def test_assert_step_not_match_should_pass_if_the_step_doesnt_match_any(mocker):
    # GIVEN
    step = mocker.MagicMock(name="Step", spec=Step, text="Step", keyword="")
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    matcher_mock = mocker.patch("radish.step_testing.matcher.matcher")
    matcher_mock.match_step.side_effect = StepImplementationNotFoundError(step)

    # WHEN
    step_not_found = assert_step_not_match(step, "some_step_func", step_registry)

    # THEN
    assert step_not_found is True


def test_assert_step_not_match_should_pass_if_the_step_doesnt_match_correct_step(
    mocker,
):
    # GIVEN
    def some_step_impl(step):
        pass

    step = mocker.MagicMock(name="Step")
    step.step_impl.func = some_step_impl
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    mocker.patch("radish.step_testing.matcher.matcher")

    # WHEN
    step_not_found = assert_step_not_match(step, "some_step_func", step_registry)

    # THEN
    assert step_not_found is True


def test_assert_step_not_match_should_fail_if_matches_correct_step_impl(mocker):
    # GIVEN
    def some_step_impl(step):
        pass

    step = mocker.MagicMock(name="Step")
    step.step_impl.func = some_step_impl
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    mocker.patch("radish.step_testing.matcher.matcher")

    # WHEN
    step_not_found = assert_step_not_match(step, "some_step_impl", step_registry)

    # THEN
    assert step_not_found is False


def test_assert_step_match_should_pass_if_step_matches(mocker):
    # GIVEN
    def some_step_impl(step):
        pass

    step = mocker.MagicMock(name="Step")
    step.step_impl.func = some_step_impl
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    mocker.patch("radish.step_testing.matcher.matcher")

    # WHEN
    step_match = assert_step_match(step, "some_step_impl", [], step_registry)

    # THEN
    assert step_match is True


def test_assert_step_match_should_fail_if_step_not_matches_any(mocker):
    # GIVEN
    step = mocker.MagicMock(name="Step")
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    matcher_mock = mocker.patch("radish.step_testing.matcher.matcher")
    matcher_mock.match_step.side_effect = StepImplementationNotFoundError(step)

    # WHEN
    step_match = assert_step_match(step, "some_step_impl", [], step_registry)

    # THEN
    assert step_match is False


def test_assert_step_match_should_fail_if_step_not_matches_wrong_step(mocker):
    # GIVEN
    def some_step_impl(step):
        pass

    step = mocker.MagicMock(name="Step")
    step.step_impl.func = some_step_impl
    step_registry = mocker.MagicMock(name="StepRegistry", spec=StepRegistry)

    mocker.patch("radish.step_testing.matcher.matcher")

    # WHEN
    step_match = assert_step_match(step, "another_step_impl", [], step_registry)

    # THEN
    assert step_match is False



"""
radish
~~~~~~

the root from red to green.  BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import itertools
import textwrap
from pathlib import Path

from lark import Transformer

from radish.models import (
    Background,
    ConstantTag,
    DefaultRule,
    Feature,
    PreconditionTag,
    Rule,
    Scenario,
    ScenarioLoop,
    ScenarioOutline,
    Step,
    Tag,
)
from radish.parser.errors import (
    RadishFirstStepMustUseFirstLevelKeyword,
    RadishScenarioOutlineExamplesInconsistentCellCount,
    RadishStepDataTableInconsistentCellCount,
)


class RadishGherkinTransformer(Transformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.featurefile_contents = None
        self.feature_id = None
        self.__step_id = None
        self.__scenario_id = None
        self.__step_keyword_ctx = None

    def prepare(
        self,
        language_spec,
        featurefile_path: Path,
        featurefile_contents: str,
        feature_id: int,
    ):
        """Prepare the Transformer for the next transformation"""
        self.language_spec = language_spec
        self.featurefile_path = featurefile_path
        self.featurefile_contents = featurefile_contents.splitlines(True)
        self.feature_id = feature_id
        self.__step_id = 1
        self.__scenario_id = 1

    def start(self, subtree):
        """Transform the root element for the radish AST"""
        if len(subtree) > 0:
            return subtree[0]
        return None

    def step_doc_string(self, subtree):
        """Transform the ``step_doc_string``-subtree for the radish AST"""
        startline = subtree[0].line
        endline = subtree[-1].line - 1
        lines = "".join(self.featurefile_contents[startline:endline])
        return textwrap.dedent(lines)

    def _table_cell(self, subtree):
        """Transform a Table Cell"""
        raw_cell_value = subtree[0].strip()
        # remove VBAR escape sequences
        cell_value = raw_cell_value.replace(r"\|", "|")
        return cell_value

    def _table_row(self, subtree):
        """Transform a Table Row"""
        return list(subtree)

    #: Transform the ``step_data_table_cell``-subtree for the radish AST
    step_data_table_cell = _table_cell
    #: Transform the ``step_data_table_row``-subtree for the radish AST
    step_data_table_row = _table_row

    def step_data_table(self, subtree):
        """Transform the ``step_data_table``-subtree for the radish AST"""
        # check if all rows have the same amount of cells
        table = list(subtree)
        if len({len(row) for row in table}) > 1:
            raise RadishStepDataTableInconsistentCellCount()
        return table

    def step_arguments(self, subtree):
        """Transform the ``step_arguments``-subtree for the radish AST"""
        if len(subtree) == 0:
            doc_string = None
            data_table = None
        elif len(subtree) == 2:
            doc_string, data_table = subtree
        elif isinstance(subtree[0], str):
            doc_string = subtree[0]
            data_table = None
        else:
            doc_string = None
            data_table = subtree[0]
        return doc_string, data_table

    def step(self, subtree):
        """Transform the ``step``-subtree for the radish AST"""
        keyword, text, (doc_string, data_table) = subtree

        keyword_line = keyword.line
        keyword = keyword.strip()
        if self.__step_keyword_ctx is None:
            if keyword not in self.language_spec.first_level_step_keywords:
                raise RadishFirstStepMustUseFirstLevelKeyword()

            self.__step_keyword_ctx = keyword
        else:
            if keyword in self.language_spec.first_level_step_keywords:
                if keyword != self.__step_keyword_ctx:
                    self.__step_keyword_ctx = keyword

        english_keyword = next(
            key
            for key, value in self.language_spec.keywords.items()
            if value == self.__step_keyword_ctx
        )

        step = Step(
            self.__step_id,
            english_keyword,
            keyword,
            text,
            doc_string,
            data_table,
            self.featurefile_path,
            keyword_line,
        )

        # increment step id for the next step
        self.__step_id += 1
        return step

    def scenario(self, subtree):
        """Transform the ``scenario``-subtree for the radish AST"""
        tags = list(itertools.takewhile(lambda t: isinstance(t, Tag), subtree))
        keyword = subtree[len(tags)]
        short_description, *steps = subtree[len(tags) + 1 :]

        scenario = Scenario(
            self.__scenario_id,
            keyword,
            short_description,
            tags,
            self.featurefile_path,
            short_description.line,
            steps,
        )

        # increment scenario id and reset step id for the next scenario
        self.__scenario_id += 1
        self.__step_id = 1
        self.__step_keyword_ctx = None
        return scenario

    #: Transform the ``example_cell``-subtree for the radish AST
    example_cell = _table_cell
    #: Transform the ``example_row``-subtree for the radish AST
    example_row = _table_row

    def examples(self, subtree):
        """Transform the ``examples``-subtree for the radish AST"""
        # check if all rows have the same amount of cells
        if len({len(row) for row in subtree}) > 1:
            raise RadishScenarioOutlineExamplesInconsistentCellCount()
        header, *rows = subtree
        return [dict(zip(header, row)) for row in rows]

    def scenario_outline(self, subtree):
        """Transform the ``scenario_outline``-subtree for the radish AST"""
        # consume Feature Tags
        tags = list(itertools.takewhile(lambda t: isinstance(t, Tag), subtree))
        keyword = subtree[len(tags)]
        short_description = subtree[len(tags) + 1]
        steps = list(
            itertools.takewhile(lambda s: isinstance(s, Step), subtree[len(tags) + 2 :])
        )
        examples_table = subtree[len(tags) + 2 + len(steps) :][0]

        scenario_outline = ScenarioOutline(
            self.__scenario_id,
            keyword,
            short_description,
            tags,
            self.featurefile_path,
            short_description.line,
            steps,
            examples_table,
        )

        # increment scenario id and reset step id for the next scenario
        self.__scenario_id += 1 + len(examples_table)
        self.__step_id = 1
        self.__step_keyword_ctx = None
        return scenario_outline

    def iterations(self, subtree):
        """Transform the ``scenario_loop``-subtree for the radish AST"""
        return int(subtree[0])

    def scenario_loop(self, subtree):
        """Transform the ``scenario_outline``-subtree for the radish AST"""
        # consume Feature Tags
        tags = list(itertools.takewhile(lambda t: isinstance(t, Tag), subtree))
        keyword = subtree[len(tags)]
        short_description = subtree[len(tags) + 1]
        steps = list(
            itertools.takewhile(lambda s: isinstance(s, Step), subtree[len(tags) + 2 :])
        )
        iterations = subtree[len(tags) + 2 + len(steps)]

        scenario_loop = ScenarioLoop(
            self.__scenario_id,
            keyword,
            short_description,
            tags,
            self.featurefile_path,
            short_description.line,
            steps,
            iterations,
        )

        # increment scenario id and reset step id for the next scenario
        self.__scenario_id += 1 + iterations
        self.__step_id = 1
        self.__step_keyword_ctx = None
        return scenario_loop

    def background(self, subtree):
        """Transform the ``background``-subtree for the radish AST"""
        keyword = subtree.pop(0)
        if len(subtree) == 0:
            short_description = None
            steps = []
        elif isinstance(subtree[0], Step):
            short_description = None
            steps = subtree
        else:
            short_description, *steps = subtree

        background = Background(
            keyword,
            short_description,
            self.featurefile_path,
            short_description.line if short_description else 0,
            steps,
        )
        return background

    def rule(self, subtree):
        """Transform the ``rule``-subtree for the radish AST"""
        keyword = subtree.pop(0)
        short_description = subtree[0]
        if len(subtree) > 1:
            scenarios = subtree[1:]
        else:
            scenarios = []

        rule = Rule(
            keyword,
            short_description,
            self.featurefile_path,
            short_description.line,
            scenarios,
        )

        # let the Scenarios know to which Rule they belong
        for scenario in scenarios:
            scenario.set_rule(rule)

        return rule

    def description(self, description_lines):
        """Transform the ``description``-subtree for the radish AST"""
        return list((str(line) for line in description_lines))

    def feature_body(self, subtree):
        """Transform the ``feature_body``-subtree for the radish AST"""
        description, *scenarios = subtree
        background, scenarios = self._expand_background_and_scenarios(scenarios)

        # create DefaultRule for scenarios without a Rul.
        scenarios_for_default_rule = list(
            itertools.takewhile(lambda s: not isinstance(s, Rule), scenarios)
        )
        rules = scenarios[len(scenarios_for_default_rule) :]
        if scenarios_for_default_rule:
            default_rule = DefaultRule(
                scenarios_for_default_rule[0].path,
                scenarios_for_default_rule[0].line,
                scenarios_for_default_rule,
            )

            # let the Scenarios in the DefaultRule know to which Rule they belong
            for scenario in scenarios_for_default_rule:
                scenario.set_rule(default_rule)

            rules = [default_rule] + rules

        return description, background, rules

    def feature(self, subtree):
        """Transform the ``feature``-subtree for the radish AST"""
        # consume Feature Tags
        tags = list(itertools.takewhile(lambda t: isinstance(t, Tag), subtree))
        keyword = subtree[len(tags)]
        short_description = subtree[len(tags) + 1]
        if len(subtree) > len(tags) + 2:
            description, background, rules = subtree[len(tags) + 2 :][0]
        else:
            description = None
            background = None
            rules = []

        feature = Feature(
            self.feature_id,
            keyword,
            short_description,
            description,
            tags,
            self.featurefile_path,
            short_description.line,
            background,
            rules,
            self.language_spec,
        )

        # let the Background know to which Feature it belongs to
        if background:
            background.set_feature(feature)

        # let the Rules know to which Feature they belong
        for rule in rules:
            rule.set_feature(feature)

        # add Background to all Rules
        for rule in rules:
            rule.set_background(background)

        return feature

    def tag(self, subtree):
        """Transform the ``tag``-subtree for the radish AST"""
        return subtree[0]

    feature_tag = tag
    scenario_tag = tag

    def std_tag(self, subtree):
        """Transform the ``tag``-subtree for the radish AST"""
        tag_name = subtree[0]
        tag = Tag(str(tag_name).strip(), self.featurefile_path, tag_name.line)
        return tag

    def precondition_tag(self, subtree):
        """Transform the ``precondition_tag``-subtree for the radish AST"""
        feature_filename, scenario_short_description = subtree
        tag = PreconditionTag(
            str(feature_filename),
            str(scenario_short_description),
            self.featurefile_path,
            feature_filename.line,
        )
        return tag

    def constant_tag(self, subtree):
        """Transform the ``constant_tag``-subtree for the radish AST"""
        key, value = subtree
        tag = ConstantTag(str(key), str(value), self.featurefile_path, key.line)
        return tag

    def _expand_background_and_scenarios(self, scenarios):
        """Expand the given list of Scenarios into Background and Scenarios if applicable"""
        background = None
        if scenarios:
            if isinstance(scenarios, Background):
                background = scenarios
                scenarios = []
            elif isinstance(scenarios, Scenario):
                pass
            elif isinstance(scenarios[0], Background):
                background = scenarios.pop(0)
        return background, scenarios



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import re
import functools
from datetime import timedelta

# We need an OrderedDict because Python 3.5 does not preverse
# the order in dictionaries
from collections import OrderedDict

import click

from radish.errors import RadishError
from radish.extensionregistry import extension
from radish.hookregistry import after
from radish.models.state import State


@extension
class JUnitXMLWriter:
    """Extension to report a jUnit XML file for the radish run"""

    OPTIONS = [
        click.Option(
            param_decls=("--junit-xml", "junit_xml"),
            help="Path to the jUnit XML report",
        )
    ]

    @classmethod
    def load(cls, config):
        if config.junit_xml:
            return cls(config.junit_xml, config.tag_expression, config.scenario_ids)
        else:
            return None

    def __init__(self, junit_xml_path, tag_expression, scenario_ids):
        try:
            from lxml import etree  # noqa
        except ImportError:
            raise RadishError(
                "if you want to use the JUnit xml writer you have to "
                "'pip install lxml'"
            )

        after.all(order=500)(
            functools.partial(
                generate_junit_xml, junit_xml_path, tag_expression, scenario_ids
            )
        )


def generate_junit_xml(junit_xml_path, tag_expression, scenario_ids, features):
    from lxml import etree

    end_states = {State.PASSED, State.FAILED}
    total_duration = sum(
        (f.duration() for f in features if f.state in end_states), timedelta()
    )

    testsuites_element = etree.Element(
        "testsuites",
        OrderedDict(
            [
                ("name", "radish"),
                ("time", "{:.3f}".format(total_duration.total_seconds())),
            ]
        ),
    )

    for feature in (f for f in features if f.has_to_run(tag_expression, scenario_ids)):
        testsuite_states = {"failures": 0, "errors": 0, "skipped": 0, "tests": 0}

        for rule in feature.rules:
            # generate test suites stats
            for scenario in (
                s for s in rule.scenarios if s.has_to_run(tag_expression, scenario_ids)
            ):
                testsuite_states["tests"] += 1
                if scenario.state in [State.UNTESTED, State.PENDING, State.SKIPPED]:
                    testsuite_states["skipped"] += 1
                if scenario.state is State.FAILED:
                    testsuite_states["failures"] += 1

            testsuite_element = etree.Element(
                "testsuite",
                OrderedDict(
                    [
                        ("name", feature.short_description),
                        ("tests", str(testsuite_states["tests"])),
                        ("skipped", str(testsuite_states["skipped"])),
                        ("failures", str(testsuite_states["failures"])),
                        ("errors", str(testsuite_states["errors"])),
                        ("time", "{:.3f}".format(feature.duration().total_seconds())),
                    ]
                ),
            )

            for scenario in (
                s for s in rule.scenarios if s.has_to_run(tag_expression, scenario_ids)
            ):
                testcase_element = etree.Element(
                    "testcase",
                    OrderedDict(
                        [
                            ("classname", feature.short_description),
                            ("name", scenario.short_description),
                            (
                                "time",
                                "{:.3f}".format(scenario.duration().total_seconds()),
                            ),
                        ]
                    ),
                )

                if scenario.state in [State.UNTESTED, State.PENDING, State.SKIPPED]:
                    skipped_element = etree.Element("skipped")
                    testcase_element.append(skipped_element)

                if scenario.state is State.FAILED:
                    steps_text = []
                    steps = scenario.steps
                    if scenario.background:
                        steps = scenario.background.steps + steps

                    for step in steps:
                        step_line = "{} {}".format(step.keyword, step.text)
                        steps_text.append(step_line)
                        if step.state is State.FAILED:
                            failure_element = etree.Element(
                                "failure",
                                OrderedDict(
                                    [
                                        ("type", step.failure_report.name),
                                        ("message", step_line),
                                    ]
                                ),
                            )
                            failure_element.text = etree.CDATA(
                                "{}\n\n{}".format(
                                    "\n".join(steps_text),
                                    _strip_ansi(step.failure_report.traceback),
                                )
                            )
                            testcase_element.append(failure_element)
                testsuite_element.append(testcase_element)
            testsuites_element.append(testsuite_element)

    with open(junit_xml_path, "wb+") as f:
        content = etree.tostring(
            testsuites_element,
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",
        )
        f.write(content)


def _strip_ansi(text):
    pattern = re.compile(r"(\\033\[\d+(?:;\d+)*m)")
    return pattern.sub("", text)



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

from collections import defaultdict


class StepImpl:
    """Represents a Step Implementation registered at the ``StepRegistry``

    A Step Implementation represents a function which is mapped to
    a Step text when running a Feature File.
    """

    __slots__ = ["keyword", "pattern", "func"]

    def __init__(self, keyword, pattern, func):
        self.keyword = keyword
        self.pattern = pattern
        self.func = func

    def __repr__(self) -> str:
        return "<StepImpl for '{}' with keyword '{}'>".format(
            self.pattern, self.keyword
        )  # pragma: no cover

    def __hash__(self):
        return hash((self.keyword, self.pattern, self.func))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (
            self.keyword == other.keyword
            and self.pattern == other.pattern  # noqa
            and self.func == other.func  # noqa
        )


class StepRegistry:
    """The ``StepRegistry`` keeps track of all declared ``StepImpl``s.

    The registered ``StepImpl``s will be used by the ``Matcher`` to
    assign to ``Step`` for execution.
    """

    #: Holds a set of keywords which can be used to register a Step Implementation.
    #  The ``And`` and ``But`` gherkin keywords are not part of this set because
    #  those can be used in all of the contexts of the keywords below.
    #  The ``Step`` keyword is used to register Step Implementations which are
    #  applicable for all the others.
    KEYWORDS = {"Given", "When", "Then", "Step"}

    def __init__(self):
        self._step_implementations = defaultdict(list)

    def step_implementations(self, keyword=None):
        """Return a dict of all registered Step Implementations"""
        if keyword is not None:
            return (
                self._step_implementations[keyword] + self._step_implementations["Step"]
            )

        return self._step_implementations

    def register(self, keyword, pattern, func):
        """Register the given ``pattern`` with the given ``func`` as Step Implementation"""
        step_impl = StepImpl(keyword, pattern, func)
        if step_impl in self._step_implementations[keyword]:
            # NOTE: allow a Step Implementation to be registered multiple times.
            #       This can happend when one step module imports another in the same
            #       RADISH_BASEDIR.
            return

        self._step_implementations[keyword].append(step_impl)

    def create_step_decorators(self, context=None):
        """Create Step decorators for all ``KEYWORDS`` and the generic ``step`` decorator.

        The created Step decorators are injected into the given ``dict``-like ``context`` object.
        If the given ``context`` is ``None`` the Steps will be injected into ``globals()``.
        """
        if context is None:
            context = globals()

        created_decorator_names = []
        for keyword in self.KEYWORDS:

            def __create_decorator(keyword):
                def __decorator(pattern):
                    def __wrapper(func):
                        self.register(keyword, pattern, func)
                        return func

                    return __wrapper

                __decorator.__doc__ = """Decorator to assign a {keyword} Step Text Pattern to a Python function

                This Python function will be run when ever a Step in the Feature File matches the
                Step Text Pattern.
                It only matches a pattern in the ``{keyword}`` context.

                For example the following Step Implementation ...

                .. code-block:: python

                    @{keyword_func}("a simple {keyword_func} Step")
                    def some_step_implementation(step):
                        assert True, "Step failed to run"

                ... will match the following Steps in our Feature File:

                .. code-block:: gherkin

                    {keyword} a simple given Step
                    And a simple given Step

                ... but it will not match Steps using any other keywords.
                """.format(
                    keyword=keyword, keyword_func=keyword.lower()
                )

                return __decorator

            decorator_name = keyword.lower()
            context[decorator_name] = __create_decorator(keyword)
            created_decorator_names.append(decorator_name)
        return created_decorator_names


#: Holds a global instance of the StepRegistry which shall be used
#  by all modules implementing Steps.
registry = StepRegistry()
__all__ = registry.create_step_decorators()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import pytest
from tagexpressions import parse

from radish.models import ScenarioOutline, Tag


def test_scenariooutline_set_feature_on_all_examples(mocker):
    """A ScenarioOutline should forward a set Feature to all its Examples"""
    # given
    feature_mock = mocker.MagicMock(name="Feature")
    scenario = ScenarioOutline(
        1, "Scenario Outline", "My ScenarioOutline", [], None, None, [], []
    )
    first_example = mocker.MagicMock(name="First Example")
    second_example = mocker.MagicMock(name="Second Example")
    scenario.examples = [first_example, second_example]

    # when
    scenario.set_feature(feature_mock)

    # then
    assert scenario.feature is feature_mock
    first_example.set_feature.assert_called_once_with(feature_mock)
    second_example.set_feature.assert_called_once_with(feature_mock)


def test_scenariooutline_set_background_on_all_examples(mocker):
    """A ScenarioOutline should forward a set Background to all its Examples"""
    # given
    background_mock = mocker.MagicMock(name="Background")
    scenario = ScenarioOutline(
        1, "Scenario Outline", "My ScenarioOutline", [], None, None, [], []
    )
    first_example = mocker.MagicMock(name="First Example")
    second_example = mocker.MagicMock(name="Second Example")
    scenario.examples = [first_example, second_example]

    # when
    scenario.set_background(background_mock)

    # then
    assert scenario.background is not None
    first_example.set_background.assert_called_once_with(background_mock)
    second_example.set_background.assert_called_once_with(background_mock)


def test_scenariooutline_set_rule_on_all_examples(mocker):
    """A ScenarioOutline should forward a set Rule to all its Examples"""
    # given
    rule_mock = mocker.MagicMock(name="Rule")
    scenario = ScenarioOutline(
        1, "Scenario Outline", "My ScenarioOutline", [], None, None, [], []
    )
    first_example = mocker.MagicMock(name="First Example")
    second_example = mocker.MagicMock(name="Second Example")
    scenario.examples = [first_example, second_example]

    # when
    scenario.set_rule(rule_mock)

    # then
    assert scenario.rule is rule_mock
    first_example.set_rule.assert_called_once_with(rule_mock)
    second_example.set_rule.assert_called_once_with(rule_mock)


def test_scenariooutline_build_examples_from_example_table():
    """A ScenarioOutline should build its Examples from the given Example Table"""
    # given & when
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "My ScenarioOutline",
        [],
        None,
        None,
        [],
        [{}, {}],  # two empty examples
    )

    # then
    assert len(scenario.examples) == 2


def test_scenariooutline_should_build_examples_with_info_in_short_description():
    """A ScenarioOutline should build its Examples with the Example Info in the short description"""
    # given & when
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "My ScenarioOutline",
        [],
        None,
        None,
        [],
        [{"foo": "bar", "bla": "meh"}, {"bar": "foo", "meh": "bla"}],
    )

    # then
    assert (
        scenario.examples[0].short_description
        == "My ScenarioOutline [foo: bar, bla: meh]"  # noqa
        # Python 3.5 has no dict ordering
        or scenario.examples[0].short_description  # noqa
        == "My ScenarioOutline [bla: meh, foo: bar]"  # noqa
    )
    assert (
        scenario.examples[1].short_description
        == "My ScenarioOutline [bar: foo, meh: bla]"  # noqa
        # Python 3.5 has no dict ordering
        or scenario.examples[1].short_description  # noqa
        == "My ScenarioOutline [meh: bla, bar: foo]"  # noqa
    )


def test_scenariooutline_should_build_examples_with_copied_steps(mocker):
    """A ScenarioOutline should build its Example with a copy of its own Steps"""
    # given & when
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "My ScenarioOutline",
        [],
        None,
        None,
        [mocker.MagicMock(name="First Step"), mocker.MagicMock(name="Second Step")],
        [{}, {}],
    )

    # then
    assert len(scenario.examples[0].steps) == 2
    assert len(scenario.examples[1].steps) == 2
    assert scenario.steps is not scenario.examples[0].steps
    assert scenario.steps is not scenario.examples[1].steps


def test_scenariooutline_should_build_examples_with_replaced_step_texts(mocker):
    """
    A ScenarioOutline should build its Example with Step Texts that have the Example Info replaced
    """
    # given & when
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "My ScenarioOutline",
        [],
        None,
        None,
        [
            mocker.MagicMock(name="First Step", text="One <foo> Three"),
            mocker.MagicMock(name="Second Step", text="Four <bar> Six"),
        ],
        [{"foo": "Two", "bar": "Five"}, {"foo": "Zwei", "bar": "Fuenf"}],
    )

    # then
    assert scenario.examples[0].steps[0].text == "One Two Three"
    assert scenario.examples[0].steps[1].text == "Four Five Six"

    assert scenario.examples[1].steps[0].text == "One Zwei Three"
    assert scenario.examples[1].steps[1].text == "Four Fuenf Six"


@pytest.mark.parametrize(
    "tagexpression, scenario_ids, expected_has_to_run",
    [
        (None, [], True),
        (parse("tag-a"), [], True),
        (parse("tag-c"), [], True),
        (parse("tag-X"), [], False),
        (None, [1], True),
        (None, [3], True),
        (None, [-1], False),
        (parse("tag-a"), [-1], False),
        (parse("tag-X"), [1], False),
        (parse("tag-X"), [3], False),
        (parse("tag-a"), [3], True),
    ],
    ids=[
        "no tagexpression, no scenario_ids => RUN",
        "tagexpression match in Scenario Tags, no scenario_ids => RUN",
        "tagexpression match in Feature Tags, no scenario_ids => RUN",
        "tagexpression no match, no scenario_ids => NO RUN",
        "no tagexpression, scenario_ids match => RUN",
        "no tagexpression, scenario_ids match Example => RUN",
        "no tagexpression, scenario_ids no match => NO RUN",
        "tagexpression match, scenario_ids no match => NO RUN",
        "tagexpression no match, scenario_ids match => NO RUN",
        "tagexpression no match, scenario_ids match Example => NO RUN",
        "tag expression match, scenario_ids match Example => RUN",
    ],
)
def test_scenario_should_correctly_evaluate_if_it_has_to_be_run(
    mocker, tagexpression, scenario_ids, expected_has_to_run
):
    """Test that a Scenario should correctly evaluate if it has to be run or not"""
    # given
    feature_mock = mocker.MagicMock(tags=[Tag("tag-c", None, None)])
    scenario = ScenarioOutline(
        1,
        "Scenario Outline",
        "My ScenarioOutline",
        [Tag("tag-a", None, None), Tag("tag-b", None, None)],
        None,
        None,
        [],
        [{"foo": "bar"}, {"foo": "meh"}],
    )
    scenario.set_feature(feature_mock)

    # when
    has_to_run = scenario.has_to_run(tagexpression, scenario_ids)

    # then
    assert has_to_run == expected_has_to_run



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""


class ExtensionRegistry:
    """The ``ExtensionRegistry`` can be used to register radish extensions

    Extensions are used to inject behavior into a radish run.
    Most common it's used for generic hooks and formatters.
    """

    def __init__(self):
        self.extensions = []

    def register(self, klass):
        """Register the extension class to the registry"""
        if klass not in self.extensions:
            self.extensions.append(klass)

    def load_extensions(self, config):
        """Load all extensions which are required by the user to load"""
        loaded_extensions = []
        for extension in self.extensions:
            load_func = getattr(extension, "load")
            loaded_extension = load_func(config)
            if loaded_extension:
                loaded_extensions.append(loaded_extension)
        return loaded_extensions

    def get_extension_options(self):
        """Return a list of all Options provided by the registered extensions"""
        options = []
        for extension in self.extensions:
            extension_options = getattr(extension, "OPTIONS", None)
            if extension_options:
                options.extend(extension_options)
        return options


def extension(klass):
    """
        Registers the class as a radish extension
    """
    registry.register(klass)
    return klass


#: Holds a global instance of the ExtensionRegistry which shall be used
#  by all modules implementing Extensions
registry = ExtensionRegistry()



"""
radish
~~~~~~

The root from red to green. BDD tooling for Python.

:copyright: (c) 2019 by Timo Furrer <tuxtimo@gmail.com>
:license: MIT, see LICENSE for more details.
"""

import functools
import random

import radish.matcher as matcher
from radish.errors import RadishError
from radish.models import Feature, Rule, Scenario, ScenarioLoop, ScenarioOutline, Step
from radish.models.state import State


class Runner:
    """Radish Feature Runner

    The Runner takes a list of Feature File ASTs to run.
    The Steps in the AST will be matched just before
    they are run.
    """

    def with_hooks(what):
        """Call hooks for a specific model"""

        def __decorator(func):
            def __wrapper(self, model, *args, **kwargs):
                only_formatters = self.config.dry_run_mode

                self.hook_registry.call(
                    what, "before", only_formatters, model, *args, **kwargs
                )
                try:
                    return func(self, model, *args, **kwargs)
                finally:
                    self.hook_registry.call(
                        what, "after", only_formatters, model, *args, **kwargs
                    )

            return __wrapper

        return __decorator

    def __init__(self, config, step_registry, hook_registry):
        self.config = config
        self.step_registry = step_registry
        self.hook_registry = hook_registry

    @with_hooks("all")
    def start(self, features):
        """Start the Runner"""
        success_state = State.PASSED if not self.config.wip_mode else State.FAILED
        feature_states = []
        for feature in features:
            if not feature.has_to_run(
                self.config.tag_expression, self.config.scenario_ids
            ):
                continue

            state = self.run_feature(feature)
            if state is not success_state and self.config.early_exit:
                return state

            feature_states.append(feature.state)

        return not any(state is not success_state for state in feature_states)

    @with_hooks("each_feature")
    def run_feature(self, feature: Feature):
        """Run the given Feature"""
        for rule in feature.rules:
            # check if this Rule has to be run
            if not rule.has_to_run(
                self.config.tag_expression, self.config.scenario_ids
            ):
                continue

            state = self.run_rule(rule)
            if state is not State.PASSED and self.config.early_exit:
                return state

        return State.PASSED

    @with_hooks("each_rule")
    def run_rule(self, rule: Rule):
        if self.config.shuffle_scenarios:
            scenarios = random.sample(rule.scenarios, len(rule.scenarios))
        else:
            scenarios = rule.scenarios

        for scenario in scenarios:
            # check if this Scenario has to be run
            if not scenario.has_to_run(
                self.config.tag_expression, self.config.scenario_ids
            ):
                continue

            if isinstance(scenario, (ScenarioOutline, ScenarioLoop)):
                state = self.run_scenario_container(scenario)
            else:
                state = self.run_scenario(scenario)

            if state is not State.PASSED and self.config.early_exit:
                return state

        return State.PASSED

    def run_scenario_container(self, scenario_container):
        if self.config.shuffle_scenarios:
            scenarios = random.sample(
                scenario_container.examples, len(scenario_container.examples)
            )
        else:
            scenarios = scenario_container.examples

        for scenario in scenarios:
            # check if this Scenario has to be run
            if not scenario.has_to_run(
                self.config.tag_expression, self.config.scenario_ids
            ):
                continue

            state = self.run_scenario(scenario)
            if state is not State.PASSED and self.config.early_exit:
                return state

        return State.PASSED

    @with_hooks("each_scenario")
    def run_scenario(self, scenario: Scenario):
        def __run_steps(steps):
            for step in steps:
                state = self.run_step(step)
                if self.config.dry_run_mode:
                    continue  # ignore Step states in dry run mode

                if state is State.RUNNING:
                    raise RadishError(
                        "The Step {} was still in RUNNING state after it has run".format(
                            step
                        )
                    )

                if state is State.FAILED:
                    break

        # run background steps
        if scenario.background:
            __run_steps(scenario.background.steps)

        # run precondition steps
        if scenario.preconditions:
            for precondition in scenario.preconditions:
                __run_steps(precondition.steps)

        if not scenario.background or scenario.background.state is State.PASSED:
            __run_steps(scenario.steps)

        return scenario.state

    @with_hooks("each_step")
    def run_step(self, step: Step):
        return self._run_step(step)

    def _run_step(self, step: Step):
        try:
            # match the Step with a Step Implementation
            matcher.match_step(step, self.step_registry)
        except Exception as exc:
            step.fail(exc)
        else:
            if not self.config.dry_run_mode:
                if self.config.debug_steps_mode:
                    step.debug(functools.partial(self._behave_like_runner, step))
                else:
                    step.run(functools.partial(self._behave_like_runner, step))

        return step.state

    def _behave_like_runner(self, step, step_line):
        """Wrapper function for a Step to run ``step.behave_like``"""
        behave_like_step_keyword, behave_like_step_text = step_line.split(maxsplit=1)
        behave_like_step = Step(
            step.id,
            behave_like_step_keyword,
            behave_like_step_keyword,
            behave_like_step_text,
            None,
            None,
            step.path,
            step.line,
        )
        behave_like_step.feature = step.feature
        behave_like_step.rule = step.rule
        behave_like_step.scenario = step.scenario

        return self._run_step(behave_like_step), behave_like_step
